;;
;;
;; sample implementation of eliza for NWN
;; demonstrating shadow door usage

(in-package :common-lisp-user)


;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;
;;
;; the elementary intro-to-ai pattern matcher
;;
;; pattern is a list of elements of types:
;;   - symbol (e.g. hi, Rob, |don't|, etc.)
;;   - numbered variable from 1 to 9, a la regular expressions
;; e.g. '(my name is 1) is a valid pattern
 



;; helpers

(defvar nil-bindings (list (cons t t)))

(defun variable? (elt)
  (numberp elt))

(defun elements-match? (pattern-element input-element)
  (and (not (variable? pattern-element))
       (equal pattern-element input-element)))

(defun make-new-binding (variable-name value bindings)
  (cons (cons variable-name value)
        bindings))

(defun get-binding-value (variable-name bindings)
  (cdr (assoc variable-name bindings)))

(defun binding-exists? (variable-name bindings)
  (assoc variable-name bindings))

;;
;; given a non-variable pattern and an input sequence, remove 
;; the pattern elements from the head of the sequence, and 
;; return as a list. e.g.
;;  (remove-pattern '(hi there) '(hi there rob)) => ((rob))
;;  (remove-pattern '(hi there) '(hi there)) => (nil)
;;  (remove-pattern '(hi there) '(hi rob)) => nil

(defun remove-pattern (pattern input) 
  (cond ((null pattern) (list input))
        ((or (null input)
             (not (elements-match? (first pattern) (first input))))
         nil)
        (t (remove-pattern (cdr pattern) (cdr input)))))


;; this function matches a pattern that begins with a variable
;;

(defun match (pattern input &optional (bindings nil-bindings))
  
  (let ((pattern-elt (first pattern)))
    
    (cond 
     
     ;; no more pattern? return the bindings.
     ((null pattern) bindings)
     
     ;; is the first pattern element a symbol? if so, match once, and repeat.
     ((not (variable? pattern-elt))
      (and (elements-match? pattern-elt (first input))
           (match (cdr pattern) (cdr input) bindings)))
     
     ;; is the variable bound already? if so, check if input matches the binding,
     ;; remove from input and continue.
     ((binding-exists? pattern-elt bindings)
      (let ((more-input (remove-pattern (get-binding-value pattern-elt bindings)
                                        input)))
        (and (not (null more-input))
             (match (cdr pattern) (first more-input)))))
     
     ;; if this variable terminates the pattern, just add input string
     ;; to its binding, and return success
     ((and pattern-elt (not (second pattern)))
      (make-new-binding pattern-elt input bindings))
     
     ;; ok, then let's match the variable. find position in the input list
     ;; where element-wise matching resumes, and bind the variable to 
     ;; everything up till that point.
     (t (let ((next-match-position 
               (position-if #'(lambda (i) (elements-match? (second pattern) i))
                            input)))
          
          (and (not (null next-match-position))
               (match (cdr pattern)
                      (subseq input next-match-position)
                      (make-new-binding pattern-elt
                                        (subseq input 0 next-match-position)
                                        bindings))))))))
               


;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;
;;
;; eliza


;; each eliza rule is of form ((pattern ...) response ...)
;;
(defvar **eliza-rules nil) 

;; helpers

(defun get-patterns (rule) (first rule))
(defun get-responses (rule) (cdr rule))
(defun pick-response (rule) 
  (let ((count (length (get-responses rule))))
    (nth (random count) (get-responses rule))))

;; flatten sublists
;;
(defun flatten-response (response)
  (apply #'append (mapcar #'(lambda (x) (if (listp x) x (list x))) response)))

;; invert pronouns
;;
(defun invert-pronouns (response)
  (sublis '((i . you) (you . i) (me . you) (am . are) (|i'm| . |you're|)
            (yourself . myself) (myself . yourself))
          response))

;; if the pattern begins with a non-variable, add one just in case.
;; then match the pattern.
(defun preprocess-and-match (pattern input)
  (let ((preprocessed (if (variable? (first pattern))
                          pattern
                        (cons 0 pattern))))
    (match preprocessed input)))

;; if one of the patterns matches the input, return its binding list,
;; nil otherwise
;;
(defun get-first-match (patterns input)
  (and (not (null patterns))
       (or (preprocess-and-match (car patterns) input)
           (get-first-match (cdr patterns) input))))
  
;; if rule matches the input, pick a response, substitute in bindings,
;; and return the response; or nil otherwise
;;
(defun match-rule (rule input)
  (let ((result (get-first-match (get-patterns rule) input)))
    (and result
         (sublis (invert-pronouns result) (pick-response rule)))))

;; go through the list of eliza rules, find the first one that 
;; matches and return its response.
;;
(defun match-eliza (input &optional (rules **eliza-rules))
  (and (not (null rules))
       (let ((result (match-rule (first rules) input)))
         (cond ((consp result)
                (flatten-response result))
               ((not (null result))
                result)
               (t (match-eliza input (cdr rules)))))))






;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;
;;
;; nwn interface



;; convert the standard string to a list of symbols

(defun preprocess-input-string (string)
  (let* ((str1 (remove-if 
                #'(lambda (x) (member x '(#\? #\! #\. #\, #\( #\) #\- #\: #\;)))
                string))
         (str2 (concatenate 'string "(" str1 ")")))
    (read-from-string str2)))

;; convert symbols into a string

(defun postprocess-output-list (list)
  (let* ((str1 (format nil "~A" list))
         (str2 (remove-if #'(lambda (x) (member x '( #\( #\) ))) str1)))
    
    str2))


;; do conversions and eliza

(defun do-string-eliza (string)
  (let ((result (match-eliza (preprocess-input-string string))))
    (if (consp result)
        (postprocess-output-list result)
      result)))


;; do eliza in sd-exp format. we expect the incoming data to be 
;; a speech s-exp of format 
;;   (speech <speaker-name> <action> <text>)
;; and we output an sd-exp output a list of format:
;;   (speak! ...), or
;;   (attack! ...), or 
;;   (animate! ...)

(defun do-sd-eliza (input)
  (let ((result (do-string-eliza (sd-speech-text input))))
    (cond ((equal result 'bow)
           (sd-send-animate animation-oneshot-bow))
          ((equal result 'attack)
           (sd-send-attack (sd-speech-speaker input)))
          ((stringp result)
           (sd-send-speak result))
          (t nil))))
    


;; do one iteration of nwn communication: 
;; read data, call the appropriate handler, and send back
;; a reply (if applicable).  returns nil if the player 
;; issued a "!quit" command, non-nil otherwise.

(defun do-sd-eliza-iteration ()
  (let* ((input (sd-read-sexp)))
    (cond 
                  
     ((and (sd-speech? input)
           (string= (sd-speech-text input) "!quit"))
      (sd-send-speak "[ done ]")
      (sleep 1)
      nil)
      
     ((sd-speech? input)
      (do-sd-eliza input)
      t)
     
     ;; other handlers go here; make sure to return t!
     
     (t t))))





;;
;;
;; read-eval-print loop
;;
;; run eliza until we hear "!quit" from the player

(defun shadow-door-eliza-loop ()
  (unwind-protect
      (progn 
        (sd-connect)
        (sd-write-line "@v")
        (while (do-sd-eliza-iteration)
          (format t ".")
          (sleep 1)))
    (sd-disconnect)))
    









;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;
;;
;; eliza rules
  

(setf **eliza-rules
  '(
    
    ;; politeness scripts
    
    (((hi) (hello) (good morning) (good evening))
     bow)
    
    (((how are you) (how do you do) (how you 1))
     (|i'm| okay) (fine thanks))
    
    (((goodbye) (bye))
     (later) (see ya))
    
    (((sorry) (i apologize) (forgive me))
     (|don't| apologize)
     (nothing to worry about))
    
    
    ;; attack trigger
    
    (((i hate you))
     attack)
    
    
    ;; some specific patterns
    
    (((computer) (computers))
     (do computers worry you?) 
     (what do you think about machines?)
     (why do you mention computers?)
     (what do you think machines have to do with your problems?))
    
    (((i love 1))
     (why do you love 1 ?) 
     (why are you attracted to 1 ?)
     (i do not understand love))
    
    (((eat 1))
     (what is you favorite food?) 
     (you like to eat 1 ?) 
     (describe 1 to me))
    
    (((name is))
     (|i'm| not interested in names))
  
    (((i remember 1) (i remembered 1) (i recall 1))
     (what do you remember about 1 ?)
     (what can you remember about 1 ?)
     (tell me more about 1))
    
    (((i know 1 to 2) (i learned 1 to 2))
     (so you know 1 to 2 ?)
     (was it difficult to 1 to 2 ?))
    
    (((i dreamt 1) (i dreamed 1))
     (really? 1 ?) 
     (have you ever fantasized 1 while you were awake?)
     (have you dreamt 1 before?))
    
    (((dream about 1))
     (how do you feel about 1 in reality?))
    
    (((dream 1))    
     (what does this dream suggest to you?) 
     (do you dream often?)
     (what persons appear in your dreams?)
     (don't you believe that dream has to do with your problem?))
    
    (((my mother 1))
     (who else in your family 2 ?) 
     (tell me more about your family)
     (tell me more about your mother))
    
    (((my father 1))
     (your father?) 
     (does he influence you strongly?) 
     (what else comes to mind when you think of your father?))

    
    ;; simple questions
    
    (((what do you 1))
     (|i'm| not sure i 1)
     (does it matter what i 1 ?))
    
    (((do you 1 to 2))
     (why do you want to know if i 1 to 2 ?)
     (does it matter whether i 1 to 2 ?)
     (and do you 1 to 2 ?))
    
    (((do you 1))
     (|i'm| not sure)
     (why do you want to know if i 1 ?)
     (well - do you 1 ?))
    
    (((what is 1))
     (why do you want to know what is 1 ?)
     (does it matter what is 1 ?))
    
    (((where is 1))
     (are you looking for 1 ?)
     (i |don't| know where 1 is))
    
    (((why don't you 2))
     (should you 2 yourself?)
     (do you believe i don't 2) 
     (perhaps i will 2 in good time))
    
    (((why 1))
     (do you think a lot about why 1 ?)
     (why do you think 1 ?)
     (so you wonder why 1 ?))
    
    (((what))
     (what do you mean?)
     (why do you say that?))
    
    
    ;; phatic statements
    
    (((i like 1) (i enjoy 1))
     (why do you like 1 ?) 
     (tell me more about 1) 
     (what is it about 1 that you like?))
    
    (((i want to 1))
     (why do you want to 1 ?)
     (what would it mean for you to 1 ?)
     (you like to 1 ?))
    
    (((i want 1))
     (why do you want 1 ?)
     (why do people want 1 ?)
     (|it's| only human to want 1))
    
    (((am glad 2))
     (how have i helped you to be 2) 
     (what makes you happy just now)
     (can you explain why you are suddenly 2))
    
    (((am sad))
     (i am sorry to hear you are depressed)
     (|i'm| sure |it's| not pleasant to be sad))
    
    (((1 are like 2))
     (what resemblance do you see between 1 and 2 ?))
    
    (((1 is like 2))    
     (in what way is it that 1 is like 2 ?)
     (what resemblance do you see?)
     (could there really be some connection?)
     (how?))
    
    (((alike))      
     (in what way?) 
     (what similarities are there?))
    
    (((same))       
     (what other connections do you see?))

    (((i hate to 1))
     (why do you hate to 1 ?))
      
    (((not 1 to 2) (don't 1 to 2))
     (why |don't| you 1 to 2 ?) (why is it you |don't| 1 to 2 ?))
    
    (((i hate 1) (not like 1) (don't like 1))
     (why do you feel that way towards 1 ?) 
     (what is it about 1 you hate?))
    
    (((i can't 2))    
     (maybe you could 2 now) 
     (what if you could 2 ?))
    
    (((i feel 2))     
     (do you often feel 2 ?))
    
    (((i felt 2))     
     (what other feelings do you have?)
     (how often have you felt 2 ?))
    
   
    ;; confusion scripts (if none of the above matched)
    
    (((1 is 2) (1 's 2))
     (is 1 2 ?)
     (why do you say 1 is 2 ?)
     (2 ? if you say so))
    
    (((if you 1))
     (do you think i 1 ?)
     (does it matter whether i 1 ?))
    
    (((if 1))
     (do you think that 1 ?)
     (how likely is it that 1 ?)
     (what do you think about 1 ?)
     (hmm - what if 1 ?))
    
    (((because you 1))
     (do i 1 ?)
     (do i really ?)
     (you think i 1 ?))
    
    (((because))
     (really ?)
     (tell me more))
    
    (((no))
     (no?)
     (|that's| awfully negative of you)
     (why not?)
     (why did you say no?))
    
    (((yes))
     (yes?)
     (are you sure?)
     (why so?)
     (why do you say that?))
    
    (((maybe))
     (maybe?)
     (what does it depend on?))
    
    (((you don't 1))
     (i |don't| 1 ?)
     (why do you say that?)
     (is it important for you to believe i |don't| 1 ?)
     (you believe i |don't| 1 ?))
    
    (((i don't 1) (i do not 1))
     (why do you not 1 ?)
     (why don't you 1 ?))
    
    (((am i 1))
     (do you believe you are 1 ?)
     (you wish you were 1 ?)
     (you wish i would tell you you are 1 ?))
     
    (((i'm 1) (am 1))
     (are you 1 ?)
     (why do you think you are 1 ?)
     (so you say you're 1)
     (in what way are you 1 ?)
     (do you want to be 1 ?))
    
    (((i was 2))       
     (were you really?) 
     (perhaps i already knew you were 2)
     (why do you tell me you were 2 now?))
    
    (((was i 2))
     (what if you were 2 ?) 
     (do you think you were 2 ?)
     (what would it mean if you were 2 ?))
    
    (((are you 2))
     (why are you interested in whether i am 2 or not?)
     (would you prefer if i weren't 2 ?)
     (perhaps i am 2 in your fantasies))
    
    (((you're 1) (you are 1))
     (you think i am 1 ?)
     (why do you say |i'm| 1 ?)
     (why do you say that?))
    
    (((are 2))
     (did you think they might not be 2)
     (possibly they are 2))
    
    (((i 2 you))   
     (perhaps in your fantasy we 2 each other))
    
    (((someone))
     (can you be more specific?))
    
    (((everyone))
     (surely not everyone) 
     (can you think of anyone in particular?)
     (who for example?) 
     (you are thinking of a special person))
    
    (((always))
     (can you think of a specific example?) 
     (when?)
     (what incident are you thinking of?)
     (really always?))
    
    (((perhaps))    
     (you do not seem quite certain))
    
    (((1))               
     (very interesting) 
     (|i'm| not sure i understand you fully)
     (what does that suggest to you?) 
     (please continue) 
     (go on) 
     (do you feel strongly about discussing such things?))

     
    ))


