;;
;;
;; allegro common lisp socket interface for shadow door

;;
;;
;; the interface provides simple functionality:
;;
;; connection - structure holding server information
;; (sd-connect) - connects to the server
;; (sd-disconnect) - disconnects from the server
;; (sd-write-sexp <s-expression>) - writes a given s-expression
;; (sd-read-sexp) => <s-expression> - nonblocking, reads an incoming s-expression




(in-package :common-lisp-user)

(eval-when (compile)
  (use-package :socket)
  (use-package :cg)
  )


;;
;;
;; struct holding connection data

(defstruct connection 
  (server-name "localhost")
  (server-port 1890)
  (socket nil))

(defvar **sd (make-connection) 
  "default connection struct")

;;
;;
;; socket maintenance

(defun sd-connect (&optional (c **sd))
  (setf (connection-socket c)
    (make-socket :remote-host (connection-server-name c)
                 :remote-port (connection-server-port c))))

(defun sd-disconnect (&optional (c **sd))
  (shutdown (connection-socket c) :direction :input)
  (shutdown (connection-socket c) :direction :output)
  (setf (connection-socket c) nil))


;;
;;
;; read and write lines

(defun sd-read-line-nonblocking (&optional (c **sd))
  (let ((s (connection-socket c))
        (list nil))
    (do ((char (read-char-no-hang s) (read-char-no-hang s)))
        ((null char) 
         (if (null list) list (concatenate 'string (reverse list))))
      (push char list))))

(defun sd-read-sexp ()
  (let ((sdexp (sd-read-line-nonblocking)))
    (and sdexp (sdexp-to-sexp sdexp))))

(defun sd-write-line (line &optional (c **sd))
  (format (connection-socket c) line)
  (finish-output (connection-socket c)))

(defun sd-write-sexp (sexp)
  (sd-write-line (sexp-to-sdexp sexp)))



;; expression conversion routines
;;
;; sd-exp (shadow door expression) is an s-exp formatted into a string, 
;; with elements all collapsed into symbols and separated by # signs,
;; and parentheses removed
;; e.g. (foo "bar") => "foo#bar"

(defun sexp-to-sdexp (sexp)
  (let* ((e1 (format nil "~S" (mapcar #'string sexp)))
         (e2 (replace-regexp e1 "\" \"" "#"))
         (e3 (replace-regexp e2 "(\"" ""))
         (final (replace-regexp e3 "\")" "")))
    final))

;; converted sd-expression becomes a list of constituent strings
;; e.g. "foo#bar" => '("foo" "bar")

(defun sdexp-to-sexp (sd-string)
  (let* ((e1 (replace-regexp sd-string "^" "(\""))
         (e2 (replace-regexp e1 "\\(.\\)$" "\\1\")"))
         (final (replace-regexp e2 "#" "\" \"")))
    (read-from-string-safely (string-downcase final))))


;; test conversion routines

#+never
(let ((sexp '("speak!" "smiles" "i'm okay"))
      (sdexp "speak!#smiles#i'm okay"))
  (and (equal sdexp (sexp-to-sdexp sexp))
       (equal sexp (sdexp-to-sexp sdexp))
       (equal sdexp (sexp-to-sdexp (sdexp-to-sexp sdexp)))
       (equal sexp (sdexp-to-sexp (sexp-to-sdexp sexp)))))


